package pwc.taxtech.atms.vat.service.impl.report.functions;

import org.apache.commons.lang3.StringUtils;
import org.apache.poi.ss.formula.OperationEvaluationContext;
import org.apache.poi.ss.formula.eval.NumberEval;
import org.apache.poi.ss.formula.eval.ValueEval;
import org.apache.poi.ss.formula.functions.FreeRefFunction;
import pwc.taxtech.atms.common.util.SpringContextUtil;
import pwc.taxtech.atms.constant.Constant;
import pwc.taxtech.atms.constant.enums.EnumOperationType;
import pwc.taxtech.atms.constant.enums.FormulaDataSourceDetailType;
import pwc.taxtech.atms.constant.enums.KeyValueConfigResultType;
import pwc.taxtech.atms.dto.vatdto.OutputInvoiceDataSourceDto;
import pwc.taxtech.atms.dto.vatdto.OutputVATInvoiceDto;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;


/// <summary>
/// 名称：销项发票
/// 功能：从发票取数，一般取发票金额或税额汇总数
/// </summary>
/// <param name="taxRate">税率取值为17%，13%，11%，6%，5%, 3%，1.5%，若所有税率都取则为99</param>
/// <param name="invoiceType">发票类型——两种取值结果“1”专用发票和“0”除专用发票以外其他发票，若所有发票都取则为99</param>
/// <param name="amountType">取值——两种取值结果“金额”0和“税额”1</param>
/// <param name="period">会计期间（开票日期）——三种取值结果 “-1”指上期，“0”当期，“99”当前累计（从期初到目前）</param>
/// <returns></returns>
public class XXFP extends FunctionBase implements FreeRefFunction {
    public XXFP(FormulaContext formulaContext) {
        super(formulaContext);
    }

    @Override
    public ValueEval evaluate(ValueEval[] args, OperationEvaluationContext ec) {
        if (args.length < 4) {
            return null;
        }

        String taxRate = getStringParam(args[0], ec);
        int invoiceType = getIntParam(args[1], ec);
        int amountType = getIntParam(args[2], ec);
        int period = getIntParam(args[3], ec);

        String formulaExpression = "XXFP(\"" + taxRate + "\"," + invoiceType + "," + amountType + "," + period + ")";
        logger.debug(formulaExpression);

        Integer curYear = formulaContext.getYear();
        int[] yearOffset = {};
        period = getPeriod(period, yearOffset);
        curYear = curYear + yearOffset[0];

        BigDecimal taxRateVal = new BigDecimal("-1");
        if (!taxRate.endsWith("%")) {
            if (taxRate.equals("99")) {
                taxRateVal = new BigDecimal("99");
            } else {
                return NumberEval.ZERO;
            }
        } else {
            taxRate = StringUtils.removeEnd(taxRate, "%");
            try {
                taxRateVal = BigDecimal.valueOf(Double.parseDouble(taxRate));
            } catch (NumberFormatException ex) {
                return NumberEval.ZERO;
            }
            taxRateVal = taxRateVal.divide(new BigDecimal("100"), 2, BigDecimal.ROUND_HALF_UP); //保留两位小数，然后四舍五入
        }

        String dbName = SpringContextUtil.projectDao.getDbNameWithYearAndOrgID(formulaContext.getOrganizationID(), curYear);
        List<OutputVATInvoiceDto> outputInvoice = SpringContextUtil.outputVATInvoiceMapper.getVatInvoiceWithItems(dbName);

        if (invoiceType != 99) {
            outputInvoice = outputInvoice.stream()
                    .filter(a -> a.getInvoice().getInvoiceType() == invoiceType)
                    .collect(Collectors.toList());
        }

        if (taxRateVal.compareTo(new BigDecimal("99")) != 0) {
            BigDecimal finalTaxRateVal = taxRateVal;
            outputInvoice = outputInvoice.stream()
                    .filter(a -> a.getInvoiceItem().getTaxRate().compareTo(finalTaxRateVal) == 0)
                    .collect(Collectors.toList());
        }

        if (period != 99) {
            int finalPeriod = period;
            outputInvoice = outputInvoice.stream().filter(a -> a.getInvoice().getPeriodID().equals(finalPeriod)).collect(Collectors.toList());
        }

        List<OutputInvoiceDataSourceDto> dataSource = new ArrayList();
        for (OutputVATInvoiceDto x : outputInvoice) {
            OutputInvoiceDataSourceDto outputInvoiceDataSourceDto = new OutputInvoiceDataSourceDto();
            outputInvoiceDataSourceDto.setAmount(FormulaHelper.roundValue(x.getInvoiceItem().getAmount(),
                    KeyValueConfigResultType.Accounting, null, formulaContext));
            outputInvoiceDataSourceDto.setResultType(KeyValueConfigResultType.Accounting.getCode());
            outputInvoiceDataSourceDto.setTaxAmount(FormulaHelper.roundValue(x.getInvoiceItem().getAmount(),
                    KeyValueConfigResultType.Accounting, null, formulaContext));
            outputInvoiceDataSourceDto.setTaxRate(x.getInvoiceItem().getTaxRate());
            outputInvoiceDataSourceDto.setBuyerName(x.getInvoice().getBuyerName());
            outputInvoiceDataSourceDto.setInvoiceCode(x.getInvoice().getClassCode());
            outputInvoiceDataSourceDto.setInvoiceNumber(x.getInvoice().getInvoiceNumber());
            outputInvoiceDataSourceDto.setInvoiceDate(x.getInvoice().getInvoiceDate());
            outputInvoiceDataSourceDto.setInvoiceType(x.getInvoice().getInvoiceType());
            outputInvoiceDataSourceDto.setPeriod(x.getInvoice().getPeriodID());
            outputInvoiceDataSourceDto.setName(Constant.DataSourceName.OutputInvoiceDataSource);
            outputInvoiceDataSourceDto.setOperationType(EnumOperationType.Single.getCode());
            dataSource.add(outputInvoiceDataSourceDto);
        }

        if (amountType == 0) {
            double val = dataSource.stream().mapToDouble(a -> a.getAmount().doubleValue()).sum();
            Long dataSoureId = saveDataSource(ec, Collections.singletonList(dataSource),
                    FormulaDataSourceDetailType.OutputInvoiceDataSourceDto, new BigDecimal(val));
            saveFormulaBlock(period, ec, formulaExpression, new BigDecimal(val), dataSoureId);
            return new NumberEval(val);
        } else if (amountType == 1) {
            double val = dataSource.stream().mapToDouble(a -> a.getTaxAmount().doubleValue()).sum();
            Long dataSoureId = saveDataSource(ec, Collections.singletonList(dataSource),
                    FormulaDataSourceDetailType.OutputInvoiceDataSourceDto, new BigDecimal(val));
            saveFormulaBlock(period, ec, formulaExpression, new BigDecimal(val), dataSoureId);
            return new NumberEval(val);
        }
        return NumberEval.ZERO;
    }
}
